
import requests

class AtomicInformation():
    def __init__(self, atom: str) -> None:
        self.atom = atom

    def atomnr(self) -> None:
        dic = {
            'H' : 1, 'He' : 2, 'Li' : 3, 'Be' : 4, 'B' : 5, 'C' : 6, 'N' : 7, 'O' : 8,
            'F' : 9, 'Ne' : 10, 'Na' : 11, 'Mg' : 12, 'Al' : 13, 'Si' : 14, 'P' : 15,
            'S' : 16, 'Cl' : 17, 'Ar' : 18, 'K' : 19, 'Ca' : 20, 'Sc' : 21, 'Ti' : 22,
            'V' : 23, 'Cr' : 24, 'Mn':  25, 'Fe' : 26, 'Co' : 27, 'Ni' : 28, 'Cu' : 29,
            'Zn' : 30, 'Ga' : 31, 'Ge' : 32, 'As' : 33, 'Se' : 34, 'Br' : 35, 'Kr' : 36,
            'Rb' : 37, 'Sr' : 38, 'Y' : 39, 'Zr' : 40, 'Nb' : 41, 'Mo' : 42, 'Tc' : 43,
            'Ru' : 44, 'Rh' : 45, 'Pd' : 46, 'Ag' : 47, 'Cd' : 48, 'In' : 49, 'Sn' : 50,
            'Sb' : 51, 'Te' : 52, 'I' : 53, 'Xe' : 54, 'Cs' : 55, 'Ba' : 56, 'La' : 57,
            'Ce' : 58, 'Pr' : 59, 'Nd' : 60, 'Pm' : 61, 'Sm' : 62, 'Eu' : 63, 'Gd' : 64,
            'Tb' : 65, 'Dy' : 66, 'Ho' : 67, 'Er' : 68, 'Tm' : 69, 'Yb' : 70, 'Lu' : 71,
            'Hf' : 72, 'Ta' : 73, 'W' : 74, 'Re' : 75, 'Os' : 76, 'Ir' : 77, 'Pt' : 78,
            'Au' : 79, 'Hg' : 80, 'Tl':  81, 'Pb' : 82, 'Bi' : 83, 'Po' : 84, 'At' : 85,
            'Rn':  86, 'Fr' : 87, 'Ra' : 88, 'Ac' : 89, 'Th' : 90, 'Pa' : 91, 'U' : 92,
            'Np' : 93, 'Pu' : 94, 'Am' : 95, 'Cm' : 96, 'Bk' : 97, 'Cf' : 98, 'Es' : 99,
            'Fm' : 100, 'Md' : 101, 'No' : 102, 'Lr' : 103, 'Rf' : 104, 'Db' : 105,
            'Sg' : 106, 'Bh' : 107, 'Hs' : 108, 'Mt' : 109, 'Ds' : 110, 'Rg' : 111,
            'Cn' : 112, 'Nh' : 113, 'Fl' : 114, 'Mc' : 115, 'Lv' : 116, 'Ts' : 117,
            'Og' : 118
        }

        try:
            self.nr = dic[self.atom]
            return self.nr
        except KeyError:
            print(f'{self.atom} is not an atom or has not been implemented in AtomicInformation.atomnr')

    def VdW(self) -> None:
        dic = {'H' : 1.09, 'C' : 1.7, 'O' : 1.52, 'Na' : 2.27, 'S' : 1.8, 'Cl' : 1.75,
               'Cu' : 1.4, 'Pd' : 1.63, 'Ag' : 1.72, 'Sb' : 2.06, 'Pt' : 1.75, 'Au' : 1.66}
               # https://periodictable.com/Properties/A/VanDerWaalsRadius.v.html

        try:
            self.vdw = dic[self.atom]
            return self.vdw
        except KeyError:
            print(f'{self.atom} is not an atom or has not been implemented AtomicInformation.VdW')


    def polarizability(self) -> None:
        dic = {'H' : 4.50114, 'C' : 8.465, 'O' : 16.15551282, 'Cu' : 33.742,
               'Ag' : 49.9843, 'Au' : 31.04, 'Ti' : 1.254517, 'Pt' : 42.515}
               # Au - J. Phys. Chem. C, 114 (48) (2010), pp. 20870-20876
               # Ag - M. Pereiro, D. Baldomir, Structure and static response of small silver clusters to an external electric field, arXiv preprint physics, 2007, 0702238.
               # Cu - Phys. Rev. A, 99 (1) (2019), p. 012503, J. Chem. Phys., 117 (7) (2002), pp. 3208-3218, J. Chem. Phys., 120 (22) (2004), pp. 10450-10454
               # Ti - Phys. Rev. B 71, 085418
               # O  - Phys. Rev. B 71, 085418

        try:
            self.pol = dic[self.atom]
            return self.pol
        except KeyError:
            print(f'{self.atom} is not an atom or has not been implemented AtomicInformation.polarizability')


class BasisSet:
    def __init__(self):
        self.BSE = "https://www.basissetexchange.org/"

    @staticmethod
    def CheckBasisSet(Program: str, BasisSet: str) -> bool:
        """Checks if a basis set for a program exits in the Basis Set Exchange

        Args:
            Program (str): Program to check for
            BasisSet (str): Basis set to check for

        Returns:
            bool: Returns true or false
        """
        try:
            response = requests.get(f'https://www.basissetexchange.org/api/basis/{BasisSet}/format/{Program}') # Request data from BSE
        except Exception:
            raise ConnectionError("Could not connect to Basis Set Exchange: Please check your internet connection") from None

        return response.status_code == 200


    def GenerateBasisSet(self, Program: str, BasisSet: str, Atoms: list, SupressHeader: bool = False) -> str:
        atoms = set(Atoms) # Remove duplicate atoms
        parameters = {'elements': [atoms]}
        try:
            response = requests.get(f'{self.BSE}/api/basis/{BasisSet}/format/{Program}', params=parameters) # Request data from BSE
        except Exception:
            raise ConnectionError("Could not connect to Basis Set Exchange: Please check your internet connection") from None

        # Check for errors
        if response.status_code != 200:
            print(response.text)
            raise RuntimeError("Could not obtain data from Basis Set Exchange. Check error message above")

        # Format the basis set text
        BasisSet = response.text.split('\n')

        if not SupressHeader:
            Info = ''
            for i in BasisSet[:10]:
                Info += f'{i}\n'
            print(Info)

        self.basis = ''
        for i in BasisSet[12:]:
            self.basis += f'{i}\n'
        return self.basis

    def AtomBasisSet(self, Program: str, BasisSet: str, Atom: str, SupressHeader: bool = False) -> str:
        parameters = {'elements': [Atom]}
        try:
            response = requests.get(f'{self.BSE}/api/basis/{BasisSet}/format/{Program}', params=parameters) # Request data from BSE
        except Exception:
            raise ConnectionError("Could not connect to Basis Set Exchange: Please check your internet connection") from None

        # Check for errors
        if response.status_code != 200:
            print(response.text)
            raise RuntimeError("Could not obtain data from Basis Set Exchange. Check error message above")

        # Format the basis set text
        BasisSet = response.text.split('\n')

        if not SupressHeader:
            Info = ''
            for i in BasisSet[:10]:
                Info += f'{i}\n'
            print(Info)

        self.basis = ''
        for i in BasisSet[12:]:
            self.basis += f'{i}\n'
        return self.basis





import numpy as np
from ase.build import fcc110, bcc110, hcp0001, diamond100
from ase.spacegroup import crystal
import Kurt.chemical_information as ci

class NanoParticle():
    def __init__(self, nanoparticle: str) -> None:
        self.structure = nanoparticle

        Atoms = {
            'Au' : ['Au'], 'Cu' : ['Cu'], 'Ag' : ['Ag'], 'TiO2' : ['Ti', 'O'],
            'Pd' : ['Pd'], 'Pt' : ['Pt'], 'NaCl' : ['Na', 'Cl'], 'CoSb3' : ['Co', 'Sb']
        }

        Length = { # In Ångstrom
            'Cu' : 3.597, 'Pd' : 3.859, 'Ag' : 4.079, 'Pt' : 3.912, 'Au' : 4.065,
            'TiO2' : [4.6, 2.95], 'NaCl' : 5.64, 'CoSb3' : 9.04
        }

        Type = {
            'Cu' : 'FCC', 'Pd' : 'FCC', 'Ag' : 'FCC', 'Pt' : 'FCC', 'Au' : 'FCC',
            'TiO2' : 'Rutile', 'NaCl' : 'Rocksalt', 'CoSb3' : 'Skutterudite'
        }

        VdW = {
            'H' : 1.09, 'C' : 1.7, 'O' : 1.52, 'Na' : 2.27, 'S' : 1.8, 'Cl' : 1.75,
            'Cu' : 1.4, 'Pd' : 1.63, 'Ag' : 1.72, 'Sb' : 2.06, 'Pt' : 1.75,
            'Au' : 1.66, 'TiO2' : 1.52, 'NaCl' : 2.27, 'CoSb3' : 2.06, 'N' : 1.55
        }

        Errors = 0

        try:
            self.atomtypes = Atoms[self.structure]
        except KeyError:
            print(f'{self.structure} has not been implemented in Atoms dictionary')
            Errors += 1

        try:
            self.lattice_length = Length[self.structure]
        except KeyError:
            print(f'{self.structure} has not been implemented in Length dictionary')
            Errors += 1

        try:
            self.lattice_type = Type[self.structure]
        except KeyError:
            print(f'{self.structure} has not been implemented in Type dictionary')
            Errors += 1

        try:
            self.vdw = VdW[self.structure]
        except KeyError:
            print(f'{self.structure} has not been implemented in Van der Waals dictionary')
            Errors += 1

        if Errors > 0:
            print(f'There were {Errors} errors for nanoparticle {nanoparticle}')
            raise ValueError

    def setInwards(self, set_to) -> None:
        self.inwards = set_to

    def setDiameter(self, dia) -> None:
        self.diameter = dia
        self.radius = self.diameter / 2

    def setRadius(self, rad) -> None:
        self.radius = rad
        self.diameter = self.radius * 2

    def makeNanoparticle(self, diameter):
        if type(self.lattice_length) == float:
            dim = int(np.ceil(diameter/self.lattice_length)) * 2
        else:
            dim1 = int(np.ceil(diameter/self.lattice_length[0]))
            dim2 = int(np.ceil(diameter/self.lattice_length[1]))

        if self.lattice_type == 'FCC':
            self.atoms = fcc110(symbol=self.atomtypes[0], size=(dim, dim, dim), a=self.lattice_length, orthogonal=True)
        elif self.lattice_type == 'BCC':
            self.atoms = bcc110(symbol=self.atomtypes[0], size=(dim, dim, 4.5*dim), a=self.lattice_length, orthogonal=True)
        elif self.lattice_type == 'HCP':
            self.atoms = hcp0001(symbol=self.atomtypes[0], size=(dim, dim, 3*dim), a=self.lattice_length[0], c=self.lattice_length[1], orthogonal=True)
        elif self.lattice_type == 'Diamond':
            self.atoms = diamond100(symbol=self.atomtypes[0], size=(dim, dim, 3*dim), a=self.lattice_length, orthogonal=True)
        elif self.lattice_type == 'Rutile':
            a = self.lattice_length[0]
            c = self.lattice_length[1]
            self.atoms = crystal([self.atomtypes[0], self.atomtypes[1]], basis=[(0, 0, 0), (0.3, 0.3, 0.0)], spacegroup=136, cellpar=[a, a, c, 90, 90, 90], size=(2, dim1, dim2))
        elif self.lattice_type == 'Rocksalt':
            a = self.lattice_length
            self.atoms = crystal([self.atomtypes[0], self.atomtypes[1]], basis=[(0, 0, 0), (0.5, 0.5, 0.5)], spacegroup=225, cellpar=[a, a, a, 90, 90, 90], size=(2, dim1, dim2))
        elif self.lattice_type == 'Skutterudite':
            a = 9.04
            self.atoms = crystal([self.atomtypes[0], self.atomtypes[1]], basis=[(0.25, 0.25, 0.25), (0.0, 0.335, 0.158)], spacegroup=204, cellpar=[a, a, a, 90, 90, 90], size=(2, dim1, dim2))
        else:
            print(f'The crystal structure type {self.lattice_type} is not implemented')
            raise ValueError

        self.atoms_symbols = self.atoms.get_chemical_symbols()
        self.atoms_pos = self.atoms.get_positions()

        return self.atoms_symbols, self.atoms_pos

    def makeSandwich(self, molecule: list, molecule_symbols: list) -> None:

        vdW_left = ci.AtomicInformation(molecule_symbols[molecule.index_min()]).VdW()
        vdW_right = ci.AtomicInformation(molecule_symbols[molecule.index_max()]).VdW()

        vacuum_dist_left = vdW_left + self.vdw
        vacuum_dist_right = vdW_right + self.vdw

        self.distance = molecule.__len__() + vacuum_dist_left + vacuum_dist_right

        if type(self.lattice_length) == float:
            left, right, left_symbols, right_symbols = self.spherical(molecule.max(), molecule.min(), vacuum_dist_left, vacuum_dist_right)
        else:
            left, right, left_symbols, right_symbols = self.slab(molecule.max(), molecule.min(), vacuum_dist_left, vacuum_dist_right)

        self.left = left
        self.right = right
        self.left_symbols = left_symbols
        self.right_symbols = right_symbols

        return self.left, self.right, self.left_symbols, self.right_symbols

    def make_sphere(self, atom_array, sphere_radii) -> list:
        shiftx = max(atom_array[:, 0])
        shifty = max(atom_array[:, 1])
        shiftz = max(atom_array[:, 2])
        atom_array[:, 0] = atom_array[:, 0] - 0.5*shiftx
        atom_array[:, 1] = atom_array[:, 1] - 0.5*shifty
        atom_array[:, 2] = atom_array[:, 2] - 0.5*shiftz
        sphere = np.empty((0, 3))
        for i in range(len(atom_array)):
            if sum(atom_array[i, :]**2)**0.5 < sphere_radii:
                sphere = np.vstack([sphere, atom_array[i, :]])
        return sphere

    def spherical(self, mol_max, mol_min, vacuum_dist_left, vacuum_dist_right):
        sphere_coor = self.make_sphere(self.atoms_pos, self.radius)

        left_hemisphere = np.empty((0, 3))
        left_symbols = np.array([])
        for i in range(len(sphere_coor)):
            if sphere_coor[i, 0] < 0:
                left_hemisphere = np.vstack([left_hemisphere, sphere_coor[i, :]])
                left_symbols = np.append(left_symbols, self.atoms_symbols[i])
        right_hemisphere = np.copy(left_hemisphere) * np.array([-1, 1, 1])
        right_symbols = left_symbols
        if self.inwards:
            left_hemisphere[:, 0] = left_hemisphere[:, 0] + mol_max + self.radius + vacuum_dist_left
            right_hemisphere[:, 0] = right_hemisphere[:, 0] + mol_min - self.radius - vacuum_dist_right
        else:
            left_hemisphere[:, 0] = left_hemisphere[:, 0] + mol_min - vacuum_dist_left
            right_hemisphere[:, 0] = right_hemisphere[:, 0] + mol_max + vacuum_dist_right
        return left_hemisphere, right_hemisphere, left_symbols, right_symbols

    def slab(self, mol_max, mol_min, vacuum_dist_left, vacuum_dist_right):
        shifty = max(self.atoms_pos[:, 1])
        shiftz = max(self.atoms_pos[:, 2])
        self.atoms_pos[:, 1] -= 0.5*shifty
        self.atoms_pos[:, 2] -= 0.5*shiftz
        left_slab = np.copy(self.atoms_pos)
        right_slab = np.copy(self.atoms_pos) * np.array([-1, 1, 1])
        left_symbols = np.copy(self.atoms_symbols)
        right_symbols = np.copy(self.atoms_symbols)
        left_slab[:, 0] = left_slab[:, 0] + mol_min - vacuum_dist_left - 2 * self.lattice_length[0]
        right_slab[:, 0] = right_slab[:, 0] + mol_max + vacuum_dist_right + 2 * self.lattice_length[0]
        return left_slab, right_slab, left_symbols, right_symbols


class Molecule():
    def __init__(self, molecule: list) -> None:
        self.molecule = molecule

    def __len__(self) -> float:
        return self.molecule[:, 0].max() - self.molecule[:, 0].min()

    def min(self) -> float:
        return min(self.molecule[:, 0])

    def max(self) -> float:
        return max(self.molecule[:, 0])

    def index_min(self) -> float:
        return np.argmin(self.molecule[:, 0])

    def index_max(self) -> float:
        return np.argmax(self.molecule[:, 0])

    def get_rotation_matrix(self, point_for_rot, direc_vec, theta) -> None:
        # Takes a point on the rotational axis, the normalized direction vector and the angle
        # The rotation matrix is 4x4 with the fourth dimension translating the origin to (a, b, c) and back
        a, b, c = point_for_rot
        u, v, w = direc_vec
        u2, v2, w2 = u*u, v*v, w*w
        cosT = np.cos(theta)
        sinT = np.sin(theta)
        m11 = u2 + (v2 + w2) * cosT
        m12 = u*v * (1-cosT) - w*sinT
        m13 = u*w * (1-cosT) + v*sinT
        m14 = (a*(v2 + w2) - u*(b*v + c*w)) * (1-cosT) + (b*w - c*v)*sinT
        m21 = u*v * (1-cosT) + w*sinT
        m22 = v2 + (u2 + w2) * cosT
        m23 = v*w * (1-cosT) - u*sinT
        m24 = (b*(u2 + w2) - v*(a*u + c*w))*(1-cosT) + (c*u - a*w)*sinT
        m31 = u*w * (1-cosT) - v*sinT
        m32 = v*w * (1-cosT) + u*sinT
        m33 = w2 + (u2 + v2) * cosT
        m34 = (c*(u2 + v2) - w*(a*u + b*v))*(1-cosT) + (a*v - b*u)*sinT

        self.rotationMatrix = np.array([m11, m12, m13, m14, m21, m22, m23, m24, m31, m32, m33, m34])

    def rotated_point(self, point) -> None:
        #Essentially matrix multiplication
        p = np.zeros(3)
        x, y, z = point
        m11, m12, m13, m14, m21, m22, m23, m24, m31, m32, m33, m34 = self.rotationMatrix
        p[0] = m11*x + m12*y + m13*z + m14
        p[1] = m21*x + m22*y + m23*z + m24
        p[2] = m31*x + m32*y + m33*z + m34

        self.rotationElements = p

    def rotateMolecule(self, alignmentAtom) -> None:
        rotxyz = np.empty((0, 3))
        for atom in self.molecule:
            self.rotated_point(atom)
            rotxyz = np.vstack([rotxyz, np.array([self.rotationElements[0], self.rotationElements[1], self.rotationElements[2]])])
        for i in range(3):
            rotxyz[:, i] -= rotxyz[alignmentAtom, i]
        rotxyz[:, 0] -= np.amax(rotxyz[:, 0])

        self.molecule = rotxyz

    def moveMolecule(self, translation) -> None:
        self.molecule += translation





import Kurt.chemical_information as ci
import numpy as np
import json
import os


def checkBasis(program: str, basis: str) -> bool:
    """Checks whether a basis set exits for a given program

    Args:
        program (str): Program to check
        functional (str): Basis set to check

    Returns:
        bool: Returns true or false
    """
    BASE_DIR = os.path.dirname(__file__)

    with open(f'{BASE_DIR}/Info.json') as info:
        info = json.load(info)
    if basis in info[program]['basis sets']:
        return True
    elif ci.BasisSet.CheckBasisSet(basis):
        return True
    return False

def checkFunctional(program: str, functional: str) -> bool:
    """Checks whether a functional exits for a given program

    Args:
        program (str): Program to check
        functional (str): Functional to check

    Returns:
        bool: Returns true or false
    """
    BASE_DIR = os.path.dirname(__file__)

    with open(f'{BASE_DIR}/Info.json') as info:
        info = json.load(info)

    return functional in info[program]['functionals']


class xyz_to:
    def __init__(self, program: str, xyz_file: str = None, memory: int = 8, ncpus: int = 8, calculation_type: str = 'Opt') -> None:
        """xyz_to ismade to make it easy to generate input files for quantum chemistry programs

        Args:
            program (str): The program to generate the input file for
            xyz_file (str, optional): The xyz file used for the input file. Defaults to None.
        """
        self.BSE = False
        self.__base_dir__ = os.path.dirname(__file__)
        self.program = program
        self.basis = 'pc-1'
        self.method = 'DFT'
        self.functional = 'B3LYP'
        self.mem = memory
        self.ncpus = ncpus
        self.calc = calculation_type

        with open(f'{self.__base_dir__}/Info.json') as info:
            self.info = json.load(info)

        if xyz_file:
            self.readXYZ(xyz_file)

    def setBasis(self, basis: str) -> None:
        """Set basis set

        Args:
            basis (str): Basis set you wish to use
        """
        if self.checkBasis(basis):
            self.basis = basis
        else:
            print(f'''{basis} was not found in the standard library of {self.program}:
Checking Basis Set Exchange''')
            if ci.BasisSet.CheckBasisSet(self.program, basis):
                self.BSE = True
                self.basis = basis
                print('Basis set found on the Basis Set Exchange\n')
            else:
                print(f'''{basis} does not seem to be available for {self.program}:
Upper and lower case characters ARE important, so you may wish to check that they are written correctly
Exiting program''')
                exit()

    def checkBasis(self, basis: str) -> bool:
        """Checks if the basis set exists for the given program

        Args:
            basis (str): Basis set to check

        Returns:
            bool: Returns true or false
        """
        return basis.upper() in self.info[self.program.upper()]['BASIS SETS']

    def setMethod(self, method: str, functional: str = None) -> None:
        """Set method (HF, MP2, DFT and so on...)

        Args:
            method (str): Method you wish to use
            functional (str, optional): Functional incase of DFT
        """
        self.method = method.upper()
        if self.method.upper() == 'DFT':
            self.setFunctional(functional)

    def setFunctional(self, functional: str) -> None:
        """Functional to use. Only relevant if using DFT method

        Args:
            functional (str): Functional uo wish to use
        """
        if self.checkFunctional(functional):
            self.functional = functional
        else:
            print(f'''{functional} does not seem to be available for {self.program}:
Exiting program''')
            exit()

    def checkFunctional(self, functional: str) -> bool:
        """Checks if the functional exists for the program

        Args:
            functional (_type_): The functional to check

        Returns:
            bool: Returns true or false
        """
        return functional.upper() in self.info[self.program.upper()]['FUNCTIONALS']

    def readXYZ(self, xyz_file: str) -> None:
        """Reads the xyz file

        Args:
            xyz_file (str): xyz file to read
        """
        self.filename = xyz_file
        with open(xyz_file, "r") as xyz:
            self.xyz_file = xyz.readlines()

    def processXYZ(self) -> None:
        """Read through the xyz file and store the contents in self.atoms
        """
        if not self.xyz_file:
            print('''xyz file has not been loaded (read):
Exiting program''')
            exit()
        self.atoms = np.empty((len(self.xyz_file[2:]), 4), dtype=object)

        for linenumber, line in enumerate(self.xyz_file[2:]):
            self.atoms[linenumber] = np.array(line.split())

    def generateDaltonInputFileText(self, charge: int) -> str:
        """Makes the text for a Dalton input file

        Args:
            charge (int): The charge of the molecule

        Returns:
            str: Returns the file text
        """
        if not hasattr(self, 'atoms'):
            print(f'''You need to process the xyz file using the class function processXYZ before generating the file:
Exiting program''')
            exit()

        self.input_filename = self.filename.replace('.xyz', '.mol')

        unique_atoms = set(self.atoms[:,0])
        self.filetext = f'''ATOMBASIS
./{self.filename}

Atomtypes={len(unique_atoms)} Charge={charge} NoSymmetry Angstrom
'''

        for unique_atom in unique_atoms:
            count = list(self.atoms[:,0]).count(unique_atom)
            if not self.BSE:
                self.filetext += f'  {ci.AtomicInformation(unique_atom).atomnr():.4f}     {count} Bas={self.basis}\n'
            else:
                BasisSet = ci.BasisSet()
                try:
                    basis_mol = BasisSet.AtomBasisSet(self.program, self.basis, unique_atom, SupressHeader=True)
                except RuntimeError:
                    print(f'''Failed to get basis set from BSE. Please check the spelling, upper-/lowercase IS important
The problem may also be that the basis set does not exist for {unique_atom}''')
                    exit()

                BlockTypes = ['s functions', 'p functions', 'd functions', 'f functions', 'g functions', 'h functions', 'i functions', 'j functions', 'k functions']
                blocks = 0
                for j in BlockTypes:
                    if j in basis_mol:
                        blocks += 1
                Block = f'{blocks}'
                for j in BlockTypes[:blocks]:
                    Block += f' {basis_mol.count(j)}'
                self.filetext += f'  Charge={ci.AtomicInformation(unique_atom).atomnr():.4f}     Atoms={count}     Blocks={Block}\n'
                basis_mol = basis_mol.replace('H','').split('\n')[5:-2]
                basis_mol = [i for i in basis_mol if 'functions' not in i]
            for j, atom in enumerate(self.atoms[:,0]):
                if atom == unique_atom:
                    self.filetext += f'{atom: <2} {self.atoms[j,1]: >14.9} {self.atoms[j,2]: >19.9} {self.atoms[j,3]: >19.9}\n'
            if self.BSE:
                for j in basis_mol:
                    self.filetext += f'{j}\n'
        return self.filetext

    def generateGaussianInputFileText(self, charge: int) -> str:
        """Makes the text for a Gaussian input file

        Args:
            charge (int): The charge of the molecule

        Returns:
            str: Returns the file text
        """
        if not hasattr(self, 'atoms'):
            print(f'''You need to process the xyz file using the class function processXYZ before generating the file:
Exiting program''')
            exit()

        basis_name = self.basis
        if self.BSE:
            BasisSet = ci.BasisSet()
            try:
                basis_mol = BasisSet.GenerateBasisSet('gaussian94', self.basis, self.atoms[:,0], SupressHeader=True)
            except RuntimeError:
                print("Failed to get basis set from BSE. Please check the spelling, upper-/lowercase is not important")
                exit()
            basis_name = self.basis
            self.basis = "GEN"

        filename_no_ext = self.filename.replace('.xyz', '')
        self.input_filename = f'{filename_no_ext}.com'

        if charge % 2 == 0:
            multiplicity = 1
        else:
            multiplicity = 2

        if self.method.upper() == 'DFT':
            method = self.functional
        else:
            method = self.method

        self.filetext = f'''%chk={filename_no_ext}.chk
%mem={self.mem}GB
%nprocshared={self.ncpus}
# {self.calc} {method}/{self.basis}

'''
        if self.BSE:
            self.filetext += f'{filename_no_ext}.xyz - {basis_name}\n'
        else:
            self.filetext += f'{filename_no_ext}.xyz\n'

        self.filetext += f'\n{charge} {multiplicity}\n'

        for atom in self.atoms:
            self.filetext += f'{atom[0]} {atom[1]: >14.9} {atom[2]: >19.9} {atom[3]: >19.9}\n'

        if self.BSE:
            self.filetext += '****\n'
            self.filetext += f'{basis_mol}\n'

        self.basis = basis_name

        return self.filetext

    def writeInputfile(self):
        """Writes the file text to an input file with similar name as the xyz file
        """
        with open(self.input_filename, 'w') as input_file:
            input_file.writelines(self.filetext)




import math
import subprocess
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc
from types import FunctionType

def Forward_search_last(file: str, text: str, error: str, quiet: bool = False):
    """Searches from the beggining of the file given to the end where it returns the linenumber of the last occurence

    Args:
        file (str): The file to search in
        text (str): The text string to search for
        error (str): If no occurences were found it will print 'No [error] could be found in [file]
        err (bool, optional): Whether or not to print error message if no occurences are found. Defaults to True.

    Returns:
        (int): Linenumber of last occurence
    """
    ps1 = subprocess.run(['grep', '-nT', text, file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = subprocess.run(['tail', '-n1'], input=ps1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = out.stdout
    if len(res) == 0:
        if not quiet:
            print(f'No {error} could be found in {file}')
        return 'NaN'
    res = res.split()[0]
    res = str(res).split(':')
    return int(res[0].replace('b\'','').replace('\'','')) - 1

def Forward_search_after_last(file: str, text1: str, text2: str, lines: int, error: str, quiet: bool = False):
    """Searches from beggining of file for last occurence of [text1] and in the following [lines] after for [text2]

    Args:
        file (str): File to search in
        text1 (str): From the last occurence of this this function will search
        text2 (str): This is what will be found in the lines following [text1]
        lines (int): How many lines after [text1] should the function search for [text2]
        error (str): If no occurences were found it will print 'No [error] could be found in [file]
        err (bool, optional): Whether or not to print error message if no occurences are found. Defaults to True.. Defaults to True.

    Returns:
        (int): Linenumber of [text2] occurence
    """
    ps1 = subprocess.run(['grep', '-nTA', f'{lines}', text1, file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    ps2 = subprocess.run(['tail', '-n', f'{lines + 1}'], input=ps1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = subprocess.run(['grep', text2], input=ps2.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = out.stdout
    if len(res) == 0:
        if not quiet:
            print(f'No {error} could be found in {file}')
        return 'NaN'
    res = res.split()[0]
    res = str(res).split('-')
    return int(res[0].replace('b\'','').replace('\'','')) - 1

def Forward_search_first(file: str, text: str, error: str, quiet: bool = False):
    """Searches from beginning of file and finds the first occurence of [text]

    Args:
        file (str): File to search in
        text (str): Text to look for
        error (str): If no occurences were found it will print 'No [error] could be found in [file]
        err (bool, optional): Whether or not to print error message if no occurences are found. Defaults to True.. Defaults to True.

    Returns:
        (int): Linenumber of first occurence
    """
    out = subprocess.run(['grep', '-nTm1', text, file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = out.stdout
    if len(res) == 0:
        if not quiet:
            print(f'No {error} could be found in {file}')
        return 'NaN'
    res = res.split()[0]
    res = str(res).split(':')
    return int(res[0].replace('b\'','').replace('\'','')) - 1

def Forward_search_all(file: str, text: str, error: str, quiet: bool = False):
    """Searches from beggining of file to end of file finding all occurences of [text]

    Args:
        file (str): File to search in
        text (str): Text to look for
        error (str): If no occurences were found it will print 'No [error] could be found in [file]
        err (bool, optional): Whether or not to print error message if no occurences are found. Defaults to True.. Defaults to True.

    Returns:
        (list): List of the linenumbers of all occurences
    """
    ps1 = subprocess.run(['grep', '-nT', text, file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out = subprocess.run(['awk', '{print $1}'], input=ps1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = out.stdout
    if len(res) == 0:
        if not quiet:
            print(f'No {error} could be found in {file}')
        return 'NaN'
    res = str(res).split('\\n')
    return [int(val.replace('b\'','').replace('\'','').replace(':','')) - 1 for val in res[:-1]]

def Resize(array: list):
    """Takes an array of arrays with varying sizes and resizes them to the same size.
    This is done by appending 'NaN' to the smaller lists.
    If the first value says 'Not implemented' the remainder of the array will be filled with 'Not implemented

    Args:
        array (list): Array of arrays with varying sizes
    """
    max_size = 0
    for arr in array:
        if len(arr) > max_size:
            max_size = len(arr)

    for i, arr in enumerate(array):
        if arr == ['Not implemented']:
            array[i] *= max_size
        else:
            array[i] += ['NaN'] * (max_size - len(arr))

def CheckForOnlyNans(array: list):
    """Function for checking if an array is fille only with the value 'NaN'

    Args:
        array (list): Array that should be looked through

    Returns:
        (bool): Returns a False/True based on whether or not the array the array only consists of 'NaN'
    """
    for i in array:
        if i != 'NaN':
           return False
    return True

def Data_Extraction(infile, Needed_Values: dict, NeededArguments, quiet: bool = False, Temperature: float = 298.15):
    Extracted_values = dict()

    infile = Output_type(str(infile), NeededArguments, quiet, Temperature)

    Extract_data(quiet, Needed_Values, infile.filename, infile.extract, infile.input)  #Extracting data
    infile.extract.__delattr__('lines') #Removing the file text from memory

    dict_keys = [*infile.extract.__dict__.keys()]
    collection_dict = dict()

    for i in dict_keys[1:]: #Collecting the data in dictionaries
        collection_dict[i] =  infile.extract.__dict__[i]

    Extracted_values[infile.filename] = collection_dict

    return Extracted_values

def Extract_data(suppressed: bool, Wanted_Values: dict, infile: str, file_text: dict, input_type: str):
    for i in Wanted_Values:
        try:
            method = getattr(type(file_text),i)
            method(file_text)
        except AttributeError:
            if not(suppressed):
                print(f'{infile}: {i} has not been implemented for {input_type}')

def Check_if_Implemented(input_file: str, Set_of_values: dict, Extracted_values: dict):
    for infile in input_file:
        for key in Set_of_values:
            for val in Set_of_values[key]:
                try:
                    Extracted_values[infile][val]
                except KeyError:
                    Extracted_values[infile][val] = ['Not implemented']

def Collect_and_sort_data(input_file: str, Set_of_values: dict, Extracted_values: dict):
    Final_arrays = dict()

    for key in Set_of_values:
        for val in Set_of_values[key]:
            Final_arrays[val] = []
            for infile in input_file:
                Final_arrays[val].append(Extracted_values[infile][val])
    return Final_arrays

def Downsizing_variable_arrays(Outputs: dict, Variable_arrays: dict, count: int, Final_arrays: dict):
    for item in Variable_arrays.items():
        if item[1] > 0:
            for val in Outputs[item[0]]:
                for file in range(0,count):
                    Final_arrays[val][file] = Final_arrays[val][file][0:item[1]]

def Create_Header(Header_text: dict, Set_of_values: dict, Final_arrays: dict):
    header = ['File']
    for key in Set_of_values.keys():
        for val in Set_of_values[key]:
            if len(Final_arrays[val][0]) > 1:
                for i in range(len(Final_arrays[val][0])):
                    header.append(f'{Header_text[val]} {i+1}')
            else:
                header.append(Header_text[val])
    return header

def Fill_output_array(Set_of_values: dict, array_input: dict, count: int, Final_arrays: dict, output_array: list):
    if count == 1:
        output_array[1,0] = array_input[0][0]
    else:
        output_array[1:,0] = np.array(np.concatenate(array_input))

    col = 1
    for key in Set_of_values.keys():
        for val in Set_of_values[key]:
            output_array[1:,col:col+len(np.array(Final_arrays[val][0]))] = np.array(Final_arrays[val])
            col += len(np.array(Final_arrays[val][0]))

def UVVIS_Spectrum(t: list, l: list, f: list, k: float, sigmacm: float):
    lambda_tot = np.zeros(len(t))
    for x in range(1,len(t)):
        lambda_tot[x] = sum((k/sigmacm)*f*np.exp(-4*np.log(2)*((1/t[x]-1/l)/(1E-7*sigmacm))**2))
    return lambda_tot

def Make_uvvis_spectrum(input_file: list, suppressed: bool, UVVIS_Spectrum: FunctionType, UVVIS: str, Extracted_Values: dict, SAVE: bool = True):
    # A LOT OF PLOT SETUP
    rc('text', usetex=True)
    xlabel_font = ylabel_font = title_font = 16
    plt.rc('font', size=12) # x/y axis font size
    N=1000 # number of calculated points in curve
    NA=6.02214199*10**23 #avogadros number
    c=299792458 #speed of light
    e=1.60217662*10**(-19) #electron charge
    me=9.10938*10**(-31) #electron mass
    pi=math.pi
    epsvac=8.8541878176*10**(-12)
    sigmacm=0.4*8065.544
    k=(NA*e**2)/(np.log(10)*2*me*c**2*epsvac)*np.sqrt(np.log(2)/pi)*10**(-1)
    inv_cm_to_au = 1/219474.63068
    Save_Dict = dict()
        # PLOT SETUP DONE
    for file in input_file:
        filename = file.replace('.out','')
        plotname = f'{filename}-uvvis.{UVVIS}'
        title = filename.replace("_"," ")

        excitations = np.array([x for x in Extracted_Values[file]['exc_energies'] if x != 'NaN' and x != 'Not implemented'])
        oscillations = np.array([x for x in Extracted_Values[file]['osc_strengths'] if x != 'NaN' and x != 'Not implemented'])

        if len(excitations) == 0 and len(oscillations) == 0:
            if not(suppressed):
                print(f'Excitation energies and oscillator strengths have either not been implemented for this output type, or none were found in the file {filename}')
            continue
        elif len(excitations) == 0:
            if not(suppressed):
                print(f'Excitation energies have either not been implemented for this output type, or none were found in the file {filename}')
            continue
        elif len(oscillations) == 0:
            if not(suppressed):
                print(f'Oscillator strengths have either not been implemented for this output type, or none were found in the file {filename}')
            continue
        elif len(excitations) > len(oscillations):
            excitations = excitations[0:len(oscillations)]
        elif len(oscillations) > len(excitations):
            oscillations = oscillations[0:len(excitations)]

        excitations = 1E7/(excitations/ inv_cm_to_au)   # From a.u. to cm^-1 to nm
        span = np.linspace(min(excitations)-20, max(excitations)+20, N, endpoint=True) # exctinction coefficient (wavelength range)

        graph = UVVIS_Spectrum(span, excitations, oscillations, k, sigmacm)
        plt.title(title)
        plt.plot(span, graph)
        plt.ylim((0,max(graph)*1.2))
        plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        plt.xlabel('Wavelength $(nm)$', fontsize=xlabel_font)
        plt.ylabel('Extinction coefficient', fontsize=ylabel_font)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        plt.savefig(plotname, format=f'{UVVIS}', dpi=600)
        plt.close()

        if SAVE:
            Save_Dict[file] = [span,graph]
    if SAVE:
        np.savez('UVVIS.npz', **Save_Dict)
        print(f'UVVIS spectrum data has been saved in UVVIS.npz')

    del Save_Dict # Deletes dictionary from memory since it is no longer needed


class Output_type:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15):
        self.filename = filename

        with open(self.filename,'r') as read:
            lines = read.readlines()[:10]

        if '* O   R   C   A *' in lines[4]: #File type = ORCA
            self.extract = orca_extract(self.filename, NeededArguments, Quiet, Temperature)
            self.input = 'ORCA'

        if '*************** Dalton - An Electronic Structure Program ***************' in lines[3]:  #File type = DALTON
            self.extract = dal_extract(self.filename, NeededArguments, Quiet, Temperature)
            self.input = 'DALTON'

        if 'Gaussian, Inc.  All Rights Reserved.' in lines[6]:  #File type = GAUSSIAN
            self.extract = gaus_extract(self.filename, NeededArguments, Quiet, Temperature)
            self.input = 'GAUSSIAN'

        if '**********  LSDalton - An electronic structure program  **********' in lines[2]:    #File type = LSDALTON
            self.extract = lsdal_extract(self.filename, NeededArguments, Quiet, Temperature)
            self.input = 'LSDALTON'

        if '!                                                       VELOXCHEM                                                        !' in lines[2]:
            self.extract = velox_extract(self.filename, NeededArguments, Quiet, Temperature)
            self.input = 'VELOXCHEM'

        del lines


class Constants:
    def __init__(self) -> None:
        self.ev_to_au = 0.036749405469679
        self.inv_cm_to_au = 1/219474.63068
        self.trans_const_fac = 1.5625517342018425307E+22 #Molar value assuming 1 bar standard pressure
        self.rot_lin_const = 20.83661793 #Assuming rigid, linear rotor and T>>Rotational temperature and rotational constant in GHz
        self.rot_poly_const = 168.5837766 #Assuming rigid, polyatomic rotor and T>>Rotational temperature and rotational constant in GHz
        self.vib_const = 3.157750419E+05 #Assuming harmonic oscillator and frequency in au
        self.gas_constant = 8.31446261815324E-03 # In kJ/(mol*K)
        self.s_trans_const = 0.3160965065 #Assuming 1 bar standard pressure and molar
        self.au_to_kJmol = 2625.4996394799


class velox_extract:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15) -> None:
        self.filename = filename
        self.NeededArguments = NeededArguments
        self.quiet = Quiet
        self.T = Temperature
        self.constants = Constants()

        self.ReadFile()

        self.end = len(self.lines)

    def ReadFile(self):
        with open(self.filename, "r") as file:
            self.lines = file.readlines()

    def _Energy(self):
        linenumber = Forward_search_last(self.filename, 'Total Energy', 'final energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-2])
            return
        self.tot_energy = 'NaN'


class gaus_extract:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15):
        self.filename = filename
        self.NeededArguments = NeededArguments
        self.quiet = Quiet
        self.T = Temperature
        self.constants = Constants()

        self.ReadFile()

        self.end = len(self.lines)

    def ReadFile(self):
        with open(self.filename, "r") as file:
            self.lines = file.readlines()

    def _Energy(self):
        linenumber = Forward_search_last(self.filename, 'Sum of electronic and zero-point Energies=', 'final energy', quiet=True)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-1]) - float(self.lines[linenumber-4].split()[-2])
            return
        linenumber = Forward_search_last(self.filename, 'SCF Done:', 'final energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[4])
            return
        self.tot_energy = 'NaN'

    def _ZPV(self):
        linenumber = Forward_search_last(self.filename, 'Sum of electronic and zero-point Energies=', 'ZPV energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.zpv = float(self.lines[linenumber].split()[-1])
            return
        self.zpv = 'NaN'

    def _Dipole_moments(self):
        linenumber = Forward_search_last(self.filename, 'Electric dipole moment (input orientation):', 'dipole moments', quiet=self.quiet)
        if type(linenumber) == int:
            self.dipolex, self.dipoley, self.dipolez, self.total_dipole = float(self.lines[linenumber+4].split()[1].replace('D','E')), float(self.lines[linenumber+5].split()[1].replace('D','E')), float(self.lines[linenumber+6].split()[1].replace('D','E')), float(self.lines[linenumber+3].split()[1].replace('D','E'))
            return
        self.dipolex = self.dipoley = self.dipolez = self.total_dipole = 'NaN'

    def _Polarizabilities(self):
        linenumber = ['NaN', 'NaN', 'NaN', 'NaN']
        searchwords = [' xx ', ' yy ', ' zz ', ' iso ']
        for i in range(len(searchwords)):
            linenumber[i] = Forward_search_after_last(self.filename, 'Dipole polarizability, Alpha (input orientation).', searchwords[i], 15, 'polarizabilities', quiet=self.quiet)
        if linenumber != ['NaN', 'NaN', 'NaN', 'NaN']:
            self.polx, self.poly, self.polz, self.iso_polar = float(self.lines[linenumber[0]].split()[1].replace('D','E')), float(self.lines[linenumber[1]].split()[1].replace('D','E')), float(self.lines[linenumber[2]].split()[1].replace('D','E')), float(self.lines[linenumber[3]].split()[1].replace('D','E'))
            return
        self.polx = self.poly = self.polz = self.iso_polar = 'NaN'

    def _Frequencies(self):
        self.freq = []
        linenumbers = Forward_search_all(self.filename, 'Frequencies --', 'frequencies', quiet=self.quiet)
        if type(linenumbers) == list:
            for i in linenumbers:
                for j in self.lines[i].split()[2:]:
                    self.freq.append(float(j)* self.constants.inv_cm_to_au)
        if len(self.freq) == 0:
            self.freq = ['NaN'] * abs(self.NeededArguments['_Frequencies'])
        if len(self.freq) < self.NeededArguments['_Frequencies']:
            self.freq += ['NaN'] * (self.NeededArguments['_Frequencies'] - len(self.freq))

    def _Excitation_energies(self):
        self.exc_energies = []
        linenumber = Forward_search_last(self.filename, 'Excitation energies and oscillator strengths:', 'excitation energies', quiet=True)
        if type(linenumber) == int:
            linenumbers = Forward_search_all(self.filename, 'Excited State', 'excitation energies', quiet=self.quiet)
            linenumbers = [i for i in linenumbers if i > linenumber]
            for i in linenumbers:
                self.exc_energies.append(float(self.lines[i].split()[4])* self.constants.ev_to_au)
        if len(self.exc_energies) == 0:
            self.exc_energies = ['NaN'] * abs(self.NeededArguments['_Excitation_energies'])
        if len(self.exc_energies) < self.NeededArguments['_Excitation_energies']:
            self.exc_energies += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.exc_energies))

    def _Oscillator_strengths(self):
        self.osc_strengths = []
        linenumber = Forward_search_last(self.filename, 'Excitation energies and oscillator strengths:', 'oscillator strengths', quiet=True)
        if type(linenumber) == int:
            linenumbers = Forward_search_all(self.filename, 'Excited State', 'oscillator strengths', quiet=self.quiet)
            linenumbers = [i for i in linenumbers if i > linenumber]
            for i in linenumbers:
                self.osc_strengths.append(float(self.lines[i].split()[-1].replace('f=','')))
        if len(self.osc_strengths) == 0:
            self.osc_strengths = ['NaN'] * abs(self.NeededArguments['_Excitation_energies'])
        if len(self.osc_strengths) < self.NeededArguments['_Excitation_energies']:
            self.osc_strengths += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.osc_strengths))

    def _RotationalConsts(self):
        self.rots = []
        linenumbers = Forward_search_all(self.filename, 'Rotational constants (GHZ):', 'rotational constants', quiet=self.quiet)
        for i in self.lines[linenumbers[-1]].split()[3:]:
            self.rots.append(float(i))
        self.rots = np.array(self.rots)
        self.rots = self.rots[self.rots != 0.0]

    def _Mass(self):
        self.mass = 0.0
        linenumber = Forward_search_last(self.filename, 'Molecular mass', 'molecular mass', quiet=self.quiet)
        if type(linenumber) == int:
            self.mass = float(self.lines[linenumber].split()[2])

    def _SymmetryNumber(self):
        self.symnum = 0
        linenumber = Forward_search_last(self.filename, 'Rotational symmetry number', 'rotational symmetry number', quiet=self.quiet)
        if type(linenumber) == int:
            self.symnum = int(self.lines[linenumber].split()[-1].replace('.',''))

    def _Multiplicity(self):
        self.multi = 0
        linenumber = Forward_search_first(self.filename, 'Multiplicity', 'multiplicity', quiet=self.quiet)
        if type(linenumber) == int:
            self.multi = int(self.lines[linenumber].split()[-1])

    def _PartitionFunctions(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.qTotal = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._SymmetryNumber()
        self._Multiplicity()
        self.qT = self.constants.trans_const_fac * self.mass ** (1.5) * self.T ** (2.5)
        if len(self.rots) == 1:
            self.qR = self.constants.rot_lin_const * self.T / (self.symnum * self.rots[0])
        else:
            self.qR = self.constants.rot_poly_const * self.T ** (1.5) / ( self.symnum * np.prod(np.array(self.rots)) ** (0.5))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.qV = np.prod(1 / (1 - np.exp( - self.constants.vib_const * realfreq / self.T)))
        self.qE = self.multi #Good approximation for most closed-shell molecules
        self.qTotal = self.qT*self.qR*self.qV*self.qE

    def _Enthalpy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.enthalpy = 'NaN'
        self._RotationalConsts()
        self.E_T = 3/2 * self.T * self.constants.gas_constant
        if len(self.rots) == 1:
            self.E_R = self.T * self.constants.gas_constant
        else:
            self.E_R = 3/2 * self.T * self.constants.gas_constant
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.E_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq * (1/2 + 1 / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1)))
        self.E_e = 0 #Good approximation for most closed-shell molecules
        self.enthalpy = (self.E_T+self.E_R+self.E_V+self.constants.gas_constant *  self.T ) / self.constants.au_to_kJmol + self.tot_energy

    def _Entropy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.entropy = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._SymmetryNumber()
        self._Multiplicity()
        self.S_T = self.constants.gas_constant * np.log(self.constants.s_trans_const * self.mass ** 1.5 * self.T ** 2.5)
        if len(self.rots) == 1:
            self.S_R = self.constants.gas_constant * np.log(self.constants.rot_lin_const * self.T / (self.symnum * self.rots[0]))
        else:
            self.S_R = self.constants.gas_constant * (3/2 + np.log(self.constants.rot_poly_const * self.T ** (1.5) / ( self.symnum * np.prod(np.array(self.rots)) ** (0.5))))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.S_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq / self.T / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1) - np.log(1-np.exp(-self.constants.vib_const * realfreq /  self.T )))
        self.S_E = self.constants.gas_constant * np.log(self.multi) #Good approximation for most closed-shell molecules
        self.entropy = self.S_T+self.S_R+self.S_V+self.S_E

    def _Gibbs(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping free energy energy calculation")
            self.gibbs = 'NaN'
            return
        self.gibbs = self.enthalpy - self.T*self.entropy / self.constants.au_to_kJmol


class orca_extract:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15):
        self.filename = filename
        self.NeededArguments = NeededArguments
        self.quiet = Quiet
        self.T = Temperature
        self.constants = Constants()

        self.ReadFile()

        self.end = len(self.lines)

    def ReadFile(self):
        with open(self.filename, "r") as file:
            self.lines = file.readlines()

    def _CPUS(self):
        linenumber = Forward_search_last(self.filename, 'Sum of individual times         ...', 'CPU time')
        if type(linenumber) == int:
            self.cpu_time = float(self.lines[linenumber].split()[-2])
            linenumber2 = Forward_search_last(self.filename, 'PAL', 'cpu count')
            if type(linenumber2) == int:
                self.pr_cpu_time = int(self.lines[linenumber2].split()[3:])
                if self.NeededArguments['_CPUS'] == 's':
                    self.pr_cpu_time *= 60
                elif self.NeededArguments['_CPUS'] == 'h':
                    self.pr_cpu_time /= 60
            else:
                self.pr_cpu_time = 'NaN'
            if self.NeededArguments['_CPUS'] == 's':
                self.cpu_time *= 60
            elif self.NeededArguments['_CPUS'] == 'h':
                self.cpu_time /= 60
            return
        self.cpu_time = 'NaN'

    def _Energy(self):
        linenumber = Forward_search_last(self.filename, 'Electronic energy', 'Final energy', quiet=True)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-2])
            return
        linenumber = Forward_search_last(self.filename, 'FINAL SINGLE POINT ENERGY', 'Final energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-1])
            return
        self.tot_energy = 'NaN'

    def _ZPV(self):
        linenumber = Forward_search_last(self.filename, 'Electronic energy', 'ZPV energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.zpv = float(self.lines[linenumber].split()[-2]) + float(self.lines[linenumber+1].split()[-4])
            return
        self.zpv = 'NaN'

    def _Enthalpy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.enthalpy = 'NaN'
            return
        self._RotationalConsts()
        self.E_T = 3/2 * self.T * self.constants.gas_constant
        if len(self.rots) == 1:
            self.E_R = self.T * self.constants.gas_constant
        else:
            self.E_R = 3/2 * self.T * self.constants.gas_constant
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.E_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq * (1/2 + 1 / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1)))
        self.E_e = 0 #Good approximation for most closed-shell molecules
        self.enthalpy = (self.E_T+self.E_R+self.E_V+self.constants.gas_constant *  self.T ) / self.constants.au_to_kJmol + self.tot_energy

    def _Gibbs(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping free energy energy calculation")
            self.gibbs = 'NaN'
            return
        self.gibbs = self.enthalpy - self.T*self.entropy / self.constants.au_to_kJmol

    def _Dipole_moments(self):
        linenumber = Forward_search_last(self.filename, 'Total Dipole Moment', 'dipole moment', quiet=self.quiet)
        if type(linenumber) == int:
            self.dipolex, self.dipoley, self.dipolez, self.total_dipole = float(self.lines[linenumber].split()[-3]), float(self.lines[linenumber].split()[-2]), float(self.lines[linenumber].split()[-1]), float(self.lines[linenumber+2].split()[-1])
            return
        self.dipolex, self.dipoley, self.dipolez, self.total_dipole = 'NaN'

    def _Polarizabilities(self):
        linenumber = Forward_search_after_last(self.filename, 'THE POLARIZABILITY TENSOR', "'diagonalized tensor:'", 10, 'polarizability', quiet=self.quiet)
        if type(linenumber) == int:
            self.polx, self.poly, self.polz, self.iso_polar = float(self.lines[linenumber+1].split()[0]), float(self.lines[linenumber+1].split()[1]), float(self.lines[linenumber+1].split()[2]), float(self.lines[linenumber+7].split()[-1])
            return
        self.polx = self.poly = self.polz = self.iso_polar = 'NaN'

    def _Excitation_energies(self):
        self.exc_energies = []
        linenumbers = Forward_search_all(self.filename, 'STATE ', 'excitation energies', quiet=self.quiet)
        if type(linenumbers) == list:
            for i in linenumbers:
                self.exc_energies.append(float(self.lines[i].split()[3]))
        if len(self.exc_energies) == 0:
            self.exc_energies = ['NaN'] * abs(self.NeededArguments['_Excitation_energies']  )
        if len(self.exc_energies) < self.NeededArguments['_Excitation_energies']:
            self.exc_energies += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.exc_energies))

    def _Oscillator_strengths(self):
        self.osc_strengths = []
        linenumber = Forward_search_last(self.filename, 'ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS', 'oscillator strengths', quiet=self.quiet)
        if type(linenumber) == int:
            for i in range(len(self.exc_energies)):
                if len(self.lines[linenumber+5+i].split()) > 6:
                    self.osc_strengths.append(float(self.lines[linenumber+5+i].split()[3]))
                else:
                    self.osc_strengths.append('NaN')
        if len(self.osc_strengths) == 0:
            self.osc_strengths = ['NaN'] * abs(self.NeededArguments['_Excitation_energies'])
        if len(self.osc_strengths) < self.NeededArguments['_Excitation_energies']:
            self.osc_strengths += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.osc_strengths))

    def _Frequencies(self):
        self.freq = []
        linenumber = Forward_search_last(self.filename, "VIBRATIONAL FREQUENCIES", 'frequencies', quiet=self.quiet)
        if type(linenumber) == int:
            for j in self.lines[linenumber+7: self.end]:
                if ": " and " 0.00 " in j:
                    pass
                elif ": " in j and not " 0.00 " in j:
                    self.freq.append(float(j.split()[1])* self.constants.inv_cm_to_au)
                else:
                    break
        if len(self.freq) == 0:
            self.freq = ['NaN'] * abs(self.NeededArguments['_Frequencies'])
        if len(self.freq) < self.NeededArguments['_Frequencies']:
            self.freq += ['NaN'] * (self.NeededArguments['_Frequencies'] - len(self.freq))

    def _RotationalConsts(self):
        self.rots = []
        linenumbers = Forward_search_first(self.filename, 'Rotational constants in MHz', 'rotational constants', quiet=self.quiet)
        for i in self.lines[linenumbers].split()[-3:]:
            self.rots.append(float(i))
        self.rots = np.array(self.rots) * 1E-3
        self.rots = self.rots[self.rots != 0.0]

    def _Mass(self):
        self.mass = 0.0
        linenumber = Forward_search_last(self.filename, 'Total Mass', 'molecular mass', quiet=self.quiet)
        if type(linenumber) == int:
            self.mass = float(self.lines[linenumber].split()[-2])

    def _SymmetryNumber(self):
        self.symnum = 0
        linenumber = Forward_search_last(self.filename, 'Symmetry Number', 'rotational symmetry number', quiet=self.quiet)
        if type(linenumber) == int:
            self.symnum = int(self.lines[linenumber].split()[-1])

    def _Multiplicity(self):
        self.multi = 0
        linenumber = Forward_search_first(self.filename, 'Multiplicity', 'multiplicity', quiet=self.quiet)
        if type(linenumber) == int:
            self.multi = int(self.lines[linenumber].split()[-1])

    def _PartitionFunctions(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.qTotal = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._Multiplicity()
        self._SymmetryNumber()
        self.qT = self.constants.trans_const_fac * self.mass ** (1.5) * self.T ** (2.5)
        if len(self.rots) == 1:
            self.qR = self.constants.rot_lin_const * self.T / (self.symnum * self.rots[0])
        else:
            self.qR = self.constants.rot_poly_const * self.T ** (1.5) / ( self.symnum * np.prod(np.array(self.rots)) ** (0.5))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.qV = np.prod(1 / (1 - np.exp( - self.constants.rot_poly_const * realfreq /  self.T )))
        self.qE = self.multi #Good approximation for most closed-shell molecules
        self.qTotal = self.qT*self.qR*self.qV*self.qE

    def _Entropy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.entropy = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._Multiplicity()
        self._SymmetryNumber()
        self.S_T = self.constants.gas_constant * np.log(self.constants.s_trans_const * self.mass ** 1.5 * self.T ** 2.5)
        if len(self.rots) == 1:
            self.S_R = self.constants.gas_constant * np.log(self.constants.rot_lin_const * self.T / (self.symnum * self.rots[0]))
        else:
            self.S_R = self.constants.gas_constant * (3/2 + np.log(self.constants.rot_poly_const * self.T ** (1.5) / ( self.symnum * np.prod(np.array(self.rots)) ** (0.5))))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.S_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq / self.T / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1) - np.log(1-np.exp(-self.constants.vib_const * realfreq /  self.T )))
        self.S_E = self.constants.gas_constant * np.log(self.multi) #Good approximation for most closed-shell molecules
        self.entropy = self.S_T+self.S_R+self.S_V+self.S_E


class dal_extract:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15):
        self.filename = filename
        self.NeededArguments = NeededArguments
        self.quiet = Quiet
        self.T = Temperature
        self.constants = Constants()

        self.ReadFile()

        self.end = len(self.lines)

    def ReadFile(self):
        with open(self.filename, "r") as file:
            self.lines = file.readlines()

    def _Energy(self):
        linenumber = Forward_search_last(self.filename, 'Total .*  energy:', 'final energy', quiet=True)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-1])
            return
        linenumber = Forward_search_last(self.filename, '@    Final .* energy:', 'final energy', quiet=True)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-1])
            return
        linenumber = Forward_search_last(self.filename, '@ Energy at final geometry is', 'final energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-2])
            return
        self.tot_energy = 'NaN'

    def _ZPV(self):
        linenumber = Forward_search_last(self.filename, 'Total Molecular Energy', 'zero-point energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.zpv = float(self.lines[linenumber+5].split()[1])
            return
        self.zpv = 'NaN'

    def _Dipole_moments(self):
        linenumber = Forward_search_last(self.filename, 'Dipole moment components', 'dipole moment', quiet=self.quiet)
        if type(linenumber) == int:
            self.dipolex, self.dipoley, self.dipolez, self.total_dipole = float(self.lines[linenumber+5].split()[1]), float(self.lines[linenumber+6].split()[1]), float(self.lines[linenumber+7].split()[1]), float(self.lines[linenumber-3].split()[0])
            return
        self.dipolex = self.dipoley = self.dipolez = self.total_dipole = 'NaN'

    def _Polarizabilities(self):
        linenumber = Forward_search_last(self.filename, 'SECOND ORDER PROPERTIES', 'polarizabilities', quiet=self.quiet)
        if type(linenumber) == int:
            self.polx, self.poly, self.polz = float(self.lines[linenumber+2].split()[-1]), float(self.lines[linenumber+5].split()[-1]), float(self.lines[linenumber+7].split()[-1])
            self.iso_polar = (self.polx + self.poly + self.polz)/3.
            return
        self.polx = self.poly = self.polz = self.iso_polar = 'NaN'

    def _Excitation_energies(self):
        self.exc_energies = []
        self.exc_type = None
        linenumber = Forward_search_last(self.filename, '@  Oscillator strengths are dimensionless.', 'excitation energies', quiet=True)
        if type(linenumber) == int:
            self.exc_type = '.EXCITA'
            for i in self.lines[linenumber+5: self.end]:
                if "@ "in i:
                    self.exc_energies.append(float(i.split()[3])* self.constants.ev_to_au)
                else:
                    break
        linenumbers = Forward_search_all(self.filename, '@ Excitation energy', 'excitation energies', quiet=self.quiet)
        if type(linenumbers) == list:
            self.exc_type = 'MCTDHF'
            for i in linenumbers:
                self.exc_energies.append(float(self.lines[i].split()[-2]))
        if len(self.exc_energies) == 0:
            self.exc_energies = ['NaN'] * abs(self.NeededArguments['_Excitation_energies'])
        if len(self.exc_energies) < self.NeededArguments['_Excitation_energies']:
            self.exc_energies += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.exc_energies))

    def _Oscillator_strengths(self):
        self.osc_strengths = []
        if self.exc_type == '.EXCITA':
            linenumber = Forward_search_last(self.filename, '@  Oscillator strengths are dimensionless.', 'oscillator strengths', quiet=self.quiet)
            if type(linenumber) == int:
                for i in self.lines[linenumber+5: self.end]:
                    if "@ " in i:
                        self.osc_strengths.append(float(i.split()[-1]))
                    else:
                        break
        elif self.exc_type == 'MCTDHF':
            linenumbers = Forward_search_all(self.filename, '@ Excitation energy', 'oscillator strengths', quiet=self.quiet)
            if type(linenumbers) == list:
                for i in linenumbers:
                    osc = 0
                    for j in self.lines[i:i+15]:
                        if '@ Oscillator strength' in j:
                            osc += float(j.split()[5])**2
                    self.osc_strengths.append(osc**0.5)
        if len(self.osc_strengths) == 0:
            self.osc_strengths = ['NaN'] * len(self.exc_energies)
            return
        if len(self.osc_strengths) < self.NeededArguments['_Excitation_energies']:
            self.osc_strengths += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.osc_strengths))

    def _Frequencies(self):
        self.freq = []
        linenumber = Forward_search_last(self.filename, 'Vibrational Frequencies and IR Intensities', 'frequencies', quiet=self.quiet)
        if type(linenumber) == int:
            for i in self.lines[linenumber+7: self.end]:
                if len(i.split()) < 1:
                    break
                self.freq.append(float(i.split()[3]))
        if len(self.freq) == 0:
            self.freq = ['NaN'] * abs(self.NeededArguments['_Frequencies'])
        if len(self.freq) < self.NeededArguments['_Frequencies']:
            self.freq += ['NaN'] * (self.NeededArguments['_Frequencies'] - len(self.freq))

    def _RotationalConsts(self):
        self.rots = []
        linenumbers = Forward_search_last(self.filename, 'Rotational constants', 'rotational constants', quiet=self.quiet)
        for i in self.lines[linenumbers+7].split()[:-1]:
            self.rots.append(float(i))
        self.rots = np.array(self.rots) * 1E-3
        self.rots = self.rots[self.rots != 0.0]

    def _Mass(self):
        self.mass = 0.0
        linenumber = Forward_search_last(self.filename, 'Total mass:', 'molecular mass')
        if type(linenumber) == int:
            self.mass = float(self.lines[linenumber].split()[-2])

    #Symmetry checking not implemented by default in Dalton
    #def _SymmetryNumber(self):
    #
    #    self.symnum = 0
    #    linenumber = Forward_search_last(self.file, 'Symmetry Number', 'rotational symmetry number')
    #    if type(linenumber) == int:
    #        self.symnum = int(self.lines[linenumber].split()[-1])

    def _Multiplicity(self):
        self.multi = 0
        linenumber = Forward_search_last(self.filename, 'Spatial symmetry', 'multiplicity', quiet=self.quiet)
        if type(linenumber) == int:
            self.multi = int(self.lines[linenumber].split()[2])

    def _PartitionFunctions(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.qTotal = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._Multiplicity()
        self.qT = self.constants.trans_const_fac * self.mass ** (1.5) * self.T ** (2.5)
        #Rotational does not give the same as Dalton, due to a correction from the assymmetric top being applied: 10.1063/1.1748490
        if len(self.rots) == 1:
            self.qR = self.constants.rot_lin_const * self.T / (self.rots[0])
        else:
            self.qR = self.constants.rot_poly_const * self.T ** (1.5) / (np.prod(self.rots) ** (0.5))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.qV = np.prod(1 / (1 - np.exp( - self.constants.vib_const * realfreq /  self.T )))
        self.qE = self.multi #Good approximation for most closed-shell molecules
        self.qTotal = self.qT*self.qR*self.qV*self.qE

    def _Entropy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.entropy = 'NaN'
            return
        self._RotationalConsts()
        self._Mass()
        self._Multiplicity()
        self.S_T = self.constants.gas_constant * np.log(self.constants.s_trans_const * self.mass ** 1.5 * self.T ** 2.5)
        if len(self.rots) == 1:
            self.S_R = self.constants.gas_constant * np.log(self.constants.rot_lin_const * self.T / (self.rots[0]))
        else:
            self.S_R = self.constants.gas_constant * (3/2 + np.log(self.constants.rot_poly_const * self.T ** (1.5) / ( np.prod(self.rots) ** (0.5))))
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.S_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq / self.T / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1) - np.log(1-np.exp(-self.constants.vib_const * realfreq /  self.T )))
        self.S_E = self.constants.gas_constant * np.log(self.multi) #Good approximation for most closed-shell molecules
        self.entropy = self.S_T+self.S_R+self.S_V+self.S_E

    def _Enthalpy(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping partition function calculation")
            self.enthalpy = 'NaN'
            return
        self._RotationalConsts()
        self.E_T = 3/2 * self.T * self.constants.gas_constant
        if len(self.rots) == 1:
            self.E_R = self.T * self.constants.gas_constant
        else:
            self.E_R = 3/2 * self.T * self.constants.gas_constant
        realfreq = np.array([x for x in self.freq if x != 'NaN'])
        realfreq = realfreq[realfreq > 0.0]
        self.E_V = self.constants.gas_constant * np.sum(self.constants.vib_const * realfreq * (1/2 + 1 / (np.exp(self.constants.vib_const * realfreq /  self.T ) - 1)))
        self.E_e = 0 #Good approximation for most closed-shell molecules
        self.enthalpy = (self.E_T+self.E_R+self.E_V+self.constants.gas_constant *  self.T ) / self.constants.au_to_kJmol + self.tot_energy

    def _Gibbs(self):
        if CheckForOnlyNans(np.array(self.freq)):
            if not(self.quiet):
                print(f"No frequencies found in {self.filename}, skipping free energy energy calculation")
            self.gibbs = 'NaN'
            return
        self.gibbs = self.enthalpy - self.T*self.entropy / self.constants.au_to_kJmol


class lsdal_extract:
    def __init__(self, filename: str, NeededArguments: dict = None, Quiet: bool = False, Temperature: float = 298.15):
        self.filename = filename
        self.NeededArguments = NeededArguments
        self.quiet = Quiet
        self.T = Temperature
        self.constants = Constants()

        self.ReadFile()

        self.end = len(self.lines)

    def ReadFile(self):
        with open(self.filename, "r") as file:
            self.lines = file.readlines()

    def _Energy(self):
        linenumber = Forward_search_last(self.filename, 'ENERGY SUMMARY', 'final energy', quiet=True)
        if type(linenumber) == int:
            for i in self.lines[linenumber+3:self.end]:
                if 'E: ' in i:
                    self.tot_energy = float(i.split()[-1])
                else:
                    return
        linenumber = Forward_search_last(self.filename, 'Final .* energy:', 'final energy', quiet=self.quiet)
        if type(linenumber) == int:
            self.tot_energy = float(self.lines[linenumber].split()[-1])
            return
        self.tot_energy = 'NaN'

    def _Dipole_moments(self):
        linenumber = Forward_search_last(self.filename, 'Permanent dipole moment', 'dipole moment', quiet=self.quiet)
        if type(linenumber) == int:
            self.dipolex, self.dipoley, self.dipolez, self.total_dipole = float(self.lines[linenumber+9].split()[1]), float(self.lines[linenumber+10].split()[1]), float(self.lines[linenumber+11].split()[1]), float(self.lines[linenumber+3].split()[0])
            return
        self.dipolex = self.dipoley = self.dipolez = self.total_dipole = 'NaN'

    def _Polarizabilities(self):
        linenumber = Forward_search_last(self.filename, '*          POLARIZABILITY TENSOR RESULTS (in a.u.)          *', 'polarizability', quiet=self.quiet)
        if type(linenumber) == int:
            self.polx, self.poly, self.polz, self.iso_polar = float(self.lines[linenumber+10].split()[-3]), float(self.lines[linenumber+11].split()[-2]), float(self.lines[linenumber+12].split()[-1]), float(self.lines[linenumber+14].split()[-1])
            return
        self.polx = self.poly = self.polz = self.iso_polar = 'NaN'

    def _Excitation_energies(self):
        self.exc_energies = []
        linenumber = Forward_search_last(self.filename, '*                   ONE-PHOTON ABSORPTION RESULTS (in a.u.)                  *', 'excitation energies', quiet=self.quiet)
        if type(linenumber) == int:
            for i in range(linenumber+8,self.end):
                if len(self.lines[i].split()) < 1:
                    break
                self.exc_energies.append(float(self.lines[i].split()[0]))
        if len(self.exc_energies) == 0:
            self.exc_energies = ['NaN'] * abs(self.NeededArguments['_Excitation_energies']  )
        if len(self.exc_energies) < self.NeededArguments['_Excitation_energies']:
            self.exc_energies += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.exc_energies))

    def _Oscillator_strengths(self):
        self.osc_strengths = []
        linenumber = Forward_search_last(self.filename, '*                   ONE-PHOTON ABSORPTION RESULTS (in a.u.)                  *', 'oscillator strengths', quiet=self.quiet)
        if type(linenumber) == int:
            for i in range(len(self.exc_energies)):
                self.osc_strengths.append(float(self.lines[linenumber+8+i].split()[-1]))
        if len(self.osc_strengths) == 0:
            self.osc_strengths = ['NaN'] * abs(self.NeededArguments['_Excitation_energies'])
        if len(self.osc_strengths) < self.NeededArguments['_Excitation_energies']:
            self.osc_strengths += ['NaN'] * (self.NeededArguments['_Excitation_energies'] - len(self.osc_strengths))
